package ru.ulstu.fc.rule.service;

import com.fuzzylite.Engine;
import com.fuzzylite.activation.General;
import com.fuzzylite.defuzzifier.Centroid;
import com.fuzzylite.norm.s.Maximum;
import com.fuzzylite.norm.t.AlgebraicProduct;
import com.fuzzylite.norm.t.Minimum;
import com.fuzzylite.rule.Rule;
import com.fuzzylite.rule.RuleBlock;
import com.fuzzylite.term.Activated;
import com.fuzzylite.term.Triangle;
import com.fuzzylite.variable.InputVariable;
import com.fuzzylite.variable.OutputVariable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;

import java.util.AbstractMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;

@Service
public class FuzzyInferenceService {
    private final static Logger LOG = LoggerFactory.getLogger(FuzzyInferenceService.class);
    private final static String OUTPUT_VARIABLE_NAME = "кредит";
    private final static String RULE_TEMPLATE = "if %s is %s and %s is %s then "
            + OUTPUT_VARIABLE_NAME
            + " is %s";
    private final static String NO_RESULT = "Нет результата";

    private Map<String, List<Entry<String, Integer>>> inputFuzzyTerms = Map.of(
            "возраст",
            List.of(
                    new AbstractMap.SimpleEntry("молодой", 35),
                    new AbstractMap.SimpleEntry("средний", 60),
                    new AbstractMap.SimpleEntry("старый", 100)),
            "доход",
            List.of(
                    new AbstractMap.SimpleEntry("небольшой", 35000),
                    new AbstractMap.SimpleEntry("средний", 100000),
                    new AbstractMap.SimpleEntry("высокий", 500000)));

    private Map<String, List<Entry<String, Integer>>> outputFuzzyTerms = Map.of(
            "кредит", List.of(new AbstractMap.SimpleEntry("небольшой", 20000),
                    new AbstractMap.SimpleEntry("средний", 100000),
                    new AbstractMap.SimpleEntry("большой", 1000000)));

    private List<String> getDemoRules() {
        return List.of(
                String.format(RULE_TEMPLATE, "возраст", "молодой", "доход", "высокий", "большой"),
                String.format(RULE_TEMPLATE, "возраст", "средний", "доход", "высокий", "средний"),
                String.format(RULE_TEMPLATE, "возраст", "старый", "доход", "высокий", "средний")
        );
    }

    private List<InputVariable> getInputVariables() {
        return List.of(getInputVariable("возраст", inputFuzzyTerms.get("возраст")),
                getInputVariable("доход", inputFuzzyTerms.get("доход")));
    }

    private InputVariable getInputVariable(String name, List<Entry<String, Integer>> terms) {
        final InputVariable input = new InputVariable();
        input.setName(name);
        input.setDescription("");
        input.setRange(0, terms.get(terms.size() - 1).getValue());
        input.setEnabled(true);
        input.setLockValueInRange(false);
        double prev = 0;
        for (int i = 0; i < terms.size(); i++) {
            Triangle term = new Triangle(terms.get(i).getKey(), prev, terms.get(i).getValue());
            prev = term.getVertexB();
            input.addTerm(term);
        }
        return input;
    }

    private <T extends Enum<T>> OutputVariable getOutputVariable(String name, List<Entry<String, Integer>> terms) {
        final OutputVariable output = new OutputVariable();
        output.setName(name);
        output.setDescription("");
        output.setRange(0, terms.get(terms.size() - 1).getValue());
        output.setEnabled(true);
        output.setAggregation(new Maximum());
        output.setDefuzzifier(new Centroid(terms.get(terms.size() - 1).getValue()));
        output.setDefaultValue(Double.NaN);
        output.setLockValueInRange(false);
        double prev = 0;
        for (int i = 0; i < terms.size(); i++) {
            Triangle term = new Triangle(terms.get(i).getKey(), prev, terms.get(i).getValue());
            prev = term.getVertexB();
            output.addTerm(term);
        }
        return output;
    }

    private RuleBlock getRuleBlock(Engine engine,
                                   List<String> rules) {
        getInputVariables().forEach(engine::addInputVariable);
        engine.addOutputVariable(getOutputVariable("кредит", outputFuzzyTerms.get("кредит")));

        RuleBlock mamdani = new RuleBlock();
        mamdani.setName("mamdani");
        mamdani.setDescription("");
        mamdani.setEnabled(true);
        mamdani.setConjunction(new Minimum());
        mamdani.setImplication(new AlgebraicProduct());
        mamdani.setActivation(new General());
        rules.forEach(r -> mamdani.addRule(Rule.parse(r, engine)));
        return mamdani;
    }

    private Engine getFuzzyEngine() {
        Engine engine = new Engine();
        engine.setName("Fuzzy rules");
        engine.setDescription("");
        return engine;
    }

    private Map<String, Double> getConsequent(Engine engine, Map<String, Double> variableValues) {
        OutputVariable outputVariable = engine.getOutputVariable(OUTPUT_VARIABLE_NAME);
        for (Map.Entry<String, Double> variableValue : variableValues.entrySet()) {
            InputVariable inputVariable = engine.getInputVariable(variableValue.getKey());
            inputVariable.setValue(variableValue.getValue());
        }
        engine.process();
        if (outputVariable != null) {
            LOG.info("Output: {}", outputVariable.getValue());
        }
        return Double.isNaN(outputVariable.getValue())
                ? Map.of(NO_RESULT, 0.0)
                : outputVariable.fuzzyOutput().getTerms().stream().collect(Collectors.toMap(t -> t.getTerm().getName(), Activated::getDegree));
    }

    public Map<String, Double> getFuzzyInference() {
        Engine engine = getFuzzyEngine();
        engine.addRuleBlock(getRuleBlock(engine, getDemoRules()));
        return getConsequent(engine, Map.of("возраст", 20.0, "доход", 250000.0));
    }

}
